﻿/**
 * 数据服务
 */
'use strict'

var sysConsts = require('../lib/scripts/sysConsts.json');
var debug = require('debug')('aprac:DataService');
var PagerInfo = require('../lib/scripts/dataSource/pagerInfo.js');
var cacheManager = require('../lib/common/cacheManager.js');
var configStore = require('../lib/scripts/configStore.js');
var iPrinter = require('../lib/scripts/iPrinter.js'); // 打印封装
var request = require("request");
var apracConfig = require("../index").ApracConfig;
var signature = require('cookie-signature');

class DataService {
    openGspx(req, newPageSrc) {
        return new Promise((resolve, reject) => {
            let url = `http://127.0.0.1:${apracConfig.port}${newPageSrc}`;
            let cookie = apracConfig.sessionIdKey + "=" +
                encodeURIComponent('s:' + signature.sign(req.sessionID, apracConfig.cookieSecret));
            let cookies = req.cookies;
            for (var name in cookies) {
                cookie += ("; " + name + "=" + cookies[name]);
            }

            request({
                url: url,
                method: "GET",
                headers: {
                    "content-type": "text/html; charset=utf-8",
                    "Cookie": cookie,
                    "server-reqesut": "true"
                }
            }, (req, res, msg) => {
                resolve(msg);
            });
        })
    }

    _updatePagerInfoData(pagerInfo, pagerInfoData) {
        //保存pagerInfo中上下文参数
        let dataSource = pagerInfo.getDataSource();
        if (dataSource) {
            pagerInfoData._oldQueryParams = dataSource.queryParams == null ? null : {
                ...dataSource.queryParams
            };
            pagerInfoData._oldItemCount = dataSource._itemCount == null ? null : dataSource._itemCount;
            pagerInfoData._isFirstLoad = false;
        }
    }

    _updatePagerInfo(pagerInfo, pagerInfoData) {
        //恢复pagerInfo中上下文参数
        let dataSource = pagerInfo.getDataSource();
        if (dataSource && pagerInfoData._isFirstLoad === false) {
            dataSource.queryParams = pagerInfoData._oldQueryParams == null ? null : {
                ...pagerInfoData._oldQueryParams
            };
            dataSource._itemCount = pagerInfoData._oldItemCount == null ? null : pagerInfoData._oldItemCount;
        }
    }

    getPagerInfoData(session, pagerId) {
        var key = sysConsts.SessionPrefixPagerInfo + pagerId;
        var pagerInfoData = session[key];

        if (!pagerInfoData) {
            throw new Error("会话超时，查询失败。请刷新该页，或重新登录!");
        }
        return pagerInfoData;
    }

    async getPagerInfo(session, pagerId) {
        var pagerInfoData = this.getPagerInfoData(session, pagerId);
        var pagerInfo = cacheManager.get(pagerInfoData._cacheId); // 取出PagerInfo
        if (!pagerInfo) {
            await this.openGspx(session.req, pagerInfoData._src);
            pagerInfo = cacheManager.get(pagerInfoData._cacheId); // 取出PagerInfo
            if (!pagerInfo) {
                throw new Error("缓存失效，查询失败。请刷新该页，或重新登录!");
            } else {
                this._updatePagerInfo(pagerInfo, pagerInfoData);
            }
        } else {
            this._updatePagerInfo(pagerInfo, pagerInfoData);
        }

        return pagerInfo;
    }

    async GetPagerData(pagerId, queryParams, orders, filter, first, count) {
        let pagerInfo = await this.getPagerInfo(this.req.session, pagerId);
        let result = pagerInfo.getPagerData(queryParams, orders, filter, first, count, null);
        let pagerInfoData = this.getPagerInfoData(this.req.session, pagerId);
        this._updatePagerInfoData(pagerInfo, pagerInfoData);
        return result;
    }

    async GetPagerSummary(pagerId) {
        var pagerInfo = await this.getPagerInfo(this.req.session, pagerId);
        let result = pagerInfo.getPagerSummary();
        let pagerInfoData = this.getPagerInfoData(this.req.session, pagerId);
        this._updatePagerInfoData(pagerInfo, pagerInfoData);
        return result;
    }

    RemoveSessionData(controlIds) {
        var session = this.req.session;
        if (session == null || controlIds == null) return 0;

        controlIds.forEach(function(controlId, i, a) {
            var data = session[controlId];
            if (data != null) {
                delete session[controlId];
                return;
            }

            var key = sysConsts.SessionPrefixPagerInfo + controlId;
            var pagerInfoData = session[key];
            if (pagerInfoData != null) {
                if (pagerInfoData._createdCount > 1) {
                    pagerInfoData._createdCount--; // 异步清理session还没完成，页面已经第二次初始化，避免新页面session被清理出现session丢失的假象
                    session[key] = pagerInfoData;
                    return;
                }
                cacheManager.remove(pagerInfoData._cacheId); // 移除PagerInfo对象的缓存
            } else {
                debug(`${session.id} PagerInfo已丢失：${key}`);
            }
            delete session[key];
        });

        return 1;
    }

    SavePanelConfig(panelId, data) {
        return configStore.saveData(this.req, panelId, data, "panel");
    }

    SaveGridConfig(gridId, data) {
        return configStore.saveData(this.req, gridId, data, "grid");
    }

    SaveFormConfig(formId, data) {
        return configStore.saveData(this.req, formId, data, "form");
    }

    GetCustomConfig(key) {
        return configStore.loadData(this.req, key, "custom");
    }

    SaveCustomConfig(key, data) {
        return configStore.saveData(this.req, key, data, "custom");
    }

    // 获取云打印相关信息
    GetCloudPrinterInfo() {
        if (!iPrinter.cloudPrinterHandler) return null;
        iPrinter.cloudPrinterHandler.req = this.req; // 注入req
        return iPrinter.cloudPrinterHandler.GetCloudPrinterInfo();
    }

    //获取打印客户端列表
    GetPrintClients() {
        if (!iPrinter.cloudPrinterHandler) return null;
        iPrinter.cloudPrinterHandler.req = this.req; // 注入req
        return iPrinter.cloudPrinterHandler.GetPrintClients();
    }

    //通过报表名称获取样式列表
    GetPrintStyles(reportName, styleFomatType, productType) {
        if (!iPrinter.cloudPrinterHandler) return null;
        var info = {
            ReportName: reportName
        };
        iPrinter.cloudPrinterHandler.req = this.req; // 注入req
        return iPrinter.cloudPrinterHandler.GetPrintStyles(info, styleFomatType, productType);
    }

    //获取授权码
    GetAccessCode() {
        if (!iPrinter.cloudPrinterHandler) return null;
        iPrinter.cloudPrinterHandler.req = this.req; // 注入req
        return iPrinter.cloudPrinterHandler.GetAccessCode();
    }
}
exports.WebMethod = DataService;