﻿/*!
 * apracjs
 * Copyright(c) 2019 by wssf
 * 2019-03-14 来自客户端的缓存组件
 * 有过期时间的缓存组件，默认是存到内存中， 也可以自定义存储
 */

StorageCache = function() {
  this._storage = new LocalStorage();
  Object.defineProperty(this, '_storage', { writable: false, enumerable: false, configurable: false });
}
StorageCache.prototype = {
  set: function(key, value, options) {
    key = _checkAndWrapKeyAsString(key);
    options = _extend({ force: true }, options);

    if (value === undefined) {
      return this.remove(key);
    }
    var cacheItem = _getCacheItem(value, options);
    try {
      this._storage.setItem(key, cacheItem);
    } catch (e) {
      console.error(e);
    }
    return value;
  },

  get: function(key, options) {
    options = _extend({ defaultValue: null }, options);
    key = _checkAndWrapKeyAsString(key);

    var cacheItem = null;
    try {
      cacheItem = this._storage.getItem(key);
      if (options && options.vs && options.vs != cacheItem.vs) return options.defaultValue; // 版本发生变化
    } catch (e) {
      return options.defaultValue;
    }

    if (_isCacheItem(cacheItem)) {
      if (_checkCacheItemIfEffective(cacheItem)) {
        return cacheItem.v;
      } else {
        this.remove(key);
      }
    }
    return options.defaultValue;
  },

  remove: function(key, options) {
    key = _checkAndWrapKeyAsString(key);
    if (options && options.group) {
      return this.removeGroup(options.group);
    }
    this._storage.removeItem(key);
    return key;
  },

  removeGroup: function(group) {
    var length = this._storage.length;
    var deleteKeys = [];
    for (var i = 0; i < length; i++) {
      var key = this._storage.key(i);
      var cacheItem = null;
      try {
        cacheItem = this._storage.getItem(key);
      } catch (e) { }

      if (cacheItem !== null && cacheItem.e !== undefined && cacheItem.g !== undefined) {
        if (cacheItem.g == group) {
          deleteKeys.push(key);
        }
      }
    }

    for (var i = 0, count = deleteKeys.length; i < count; i++) {
      this.remove(deleteKeys[i]);
    }
    return deleteKeys;
  },

  removeExpires: function() {
    var length = this._storage.length;
    var deleteKeys = [];
    var _this = this;
    for (var i = 0; i < length; i++) {
      var key = this._storage.key(i);
      var cacheItem = null;
      try {
        cacheItem = this._storage.getItem(key);
      } catch (e) { }

      if (cacheItem !== null && cacheItem.e !== undefined) {
        var timeNow = (new Date()).getTime();
        if (timeNow >= cacheItem.e) {
          deleteKeys.push(key);
        }
      }
    }

    for (var i = 0, count = deleteKeys.length; i < count; i++) {
      this.remove(deleteKeys[i]);
    }
    return deleteKeys;
  },

  clear: function() {
    this._storage.clear();
  },

  add: function(key, value, options) {
    key = _checkAndWrapKeyAsString(key);
    options = _extend({ force: true }, options);

    try {
      var cacheItem = this._storage.getItem(key);
      if (!_isCacheItem(cacheItem) || !_checkCacheItemIfEffective(cacheItem)) {
        this.set(key, value, options);
        return true;
      }
    } catch (e) {
      this.set(key, value, options);
      return true;
    }
    return false;
  },

  replace: function(key, value, options) {
    key = _checkAndWrapKeyAsString(key);
    var cacheItem = null;
    try {
      cacheItem = this._storage.getItem(key);
    } catch (e) {
      return false;
    }

    if (_isCacheItem(cacheItem)) {
      if (_checkCacheItemIfEffective(cacheItem)) {
        this.set(key, value, options);
        return true;
      } else {
        this.remove(key);
      }
    }
    return false;
  },

  touch: function(key, exp) {
    key = _checkAndWrapKeyAsString(key);
    var cacheItem = null;
    try {
      cacheItem = this._storage.getItem(key);
    } catch (e) {
      return false;
    }
    if (_isCacheItem(cacheItem)) {
      if (_checkCacheItemIfEffective(cacheItem)) {
        this.set(key, this.get(key), { exp: exp });
        return true;
      } else {
        this.remove(key);
      }
    }
    return false;
  }
}


var _maxExpireDate = new Date('Fri, 31 Dec 9999 23:59:59 UTC');
var _defaultExpire = _maxExpireDate;

function _extend(obj, props) {
  for (var key in props) obj[key] = props[key];
  return obj;
}

function _isValidDate(date) {
  return Object.prototype.toString.call(date) === '[object Date]' && !isNaN(date.getTime());
}

function _getExpiresDate(expires, now) {
  now = now || new Date();

  if (typeof expires === 'number') {
    expires = expires === Infinity ? _maxExpireDate : new Date(now.getTime() + expires * 1000);
  } else if (typeof expires === 'string') {
    expires = new Date(expires);
  }

  if (expires && !_isValidDate(expires)) {
    throw new Error('`expires` 参数不能转为有效的日期类型');
  }

  return expires;
}

function _getCacheItem(value, options) {
  var exp = options.exp;
  var obj = {};
  obj.c = (new Date()).getTime();
  exp = exp || _defaultExpire;
  var expires = _getExpiresDate(exp);
  obj.e = expires.getTime();
  obj.v = value;
  if (options.group) obj.g = options.group;
  if (options.vs) obj.vs = options.vs;
  return obj;
}

function _isCacheItem(item) {
  if (typeof item !== 'object') {
    return false;
  }
  if (item) {
    if ('c' in item && 'e' in item && 'v' in item) {
      return true;
    }
  }
  return false;
}

function _checkCacheItemIfEffective(cacheItem) {
  var timeNow = (new Date()).getTime();
  return timeNow < cacheItem.e;
}

function _checkAndWrapKeyAsString(key) {
  if (typeof key !== 'string') {
    console.warn(key + ' 使用了key,但不是字符串格式.');
    key = String(key);
  }
  return key;
}

/**
 * 本地内存存储器对象
 */
class LocalStorage {
  constructor() {
    this.length = this.getCount;
    this.cache = {};
  }

  setItem(key, value) {
    this.cache[key] = value;
  }

  getItem(key) {
    return this.cache[key];
  }

  removeItem(key) {
    delete this.cache[key];
  }

  key(index) {
    var keys = Object.keys(this.cache);
    return keys[index];
  }

  clear() {
    this.cache = {};
  }

  getCount() {
    return Object.keys(this.cache).length;
  }
}

module.exports = new StorageCache();