﻿/*!
 * apracjs
 * Copyright(c) 2017 by wssf
 * 2017-06-14
 * aprac系统内部使用，不能提供给业务，否则缓存容易被业务clear或remove，或覆盖等，业务使用cache.js
 */

'use strict'
var config = require('../config.js');
var apracConfig = config.apracConfig;

// 缓存内存对象
class MemoryStore {
    constructor() {
        this.cache = {};
        this.sessionids = {};
        this.maxAge = apracConfig.sessionTimeout * 60 * 1000 + 10 * 1000; // 直接使用session过期时间配置，间隔20分钟+10s检查一次
        //this.maxAge = 5 * 1000; // 测试用，30s检查1次
        this._refreshTime();
    }

    add(key, value, sid) {
        this.cache[key] = value;

        // 按sessionid对缓存分组，定时检查session过期情况，如果session过期了，说明用户下线或没用或关闭浏览器了，就需要清理过期session下面的所有缓存
        if (sid) {
            if (!this.sessionids[sid]) {
                this.sessionids[sid] = {}; // 按sessionid创建分组
            }
            this.sessionids[sid][key] = null; // 按组记录缓存的key，如果session过期了，对应组    
        }
    }

    get(key) {
        return this.cache[key];
    }

    remove(key, sid) {
        var value = this.cache[key];
        delete this.cache[key];
        if (typeof value == 'object' && value.dispose) {
            value.dispose(); // 如果缓存的对象实现了dispose函数(PagerInfo)，则通知它清理关联的引用(PagerDataSource)，js里面如果一个对象有引用关系，就一直得不到释放。所以业务的数据源里面定义了大对象一定要置空，比如：this._dataSource=null;
        }
        // 如果是页面关闭，触发的remove，不需要管sessionids列表都可以，里面存的都是null引用，只是一个清单而已没有任何影响，待到session到期自动清理sessionids清单的
        return value;
    }

    clear() {
        this.cache = {};
    }

    checkExpires(req) {
        if (!req) return;
        var sessionStore = req.sessionStore;
        if (!sessionStore) return;

        var cur = new Date().getTime();
        if (cur < this.nextTime) return; // 还没到检查的时间，这样做的目的是减少检查session状态的频率，session是20分钟超时，那么20分钟后检查超时才有意义，不然太频繁了
        this._refreshTime();  // 刷新下次检查的时间

        try {
            var sids = Object.keys(this.sessionids);
            console.log('当前cacheManager缓存分组数:' + sids.length);
            console.log('当前cacheManager缓存keys总数:' + Object.keys(this.cache).length);

            var that = this;
            for (var i = sids.length - 1; i >= 0; i--) {
                var sid = sids[i];
                // 调用session组件获取session，它内部做了过期检测和清理的；
                sessionStore.get(sid, function(err, sess) {
                    if (err || !sess) { // 过期或没有数据
                        that._removeGroup(sid); // 如果session都过期了，那么内存里面对应Session的所有缓存的东西全部清理掉引用关系
                    }
                });
            }
        } catch (e) {
            console.log("清理缓存发生异常", e);  // 必须吃掉异常，不然影响当前请求页面
        }
    }

    _refreshTime() {
        this.nextTime = new Date().getTime() + this.maxAge; // 下一次检查的时间
    }

    _removeGroup(sid) {
        var obj = this.sessionids[sid]; // 取出分组
        if (!obj) return;
        var keys = Object.keys(obj);
        console.log(`清理过期的缓存[${sid}]，该分组一共有[${keys.length}]个缓存项目`);
        delete this.sessionids[sid]; // 删除分组

        // 删除分组里面所有缓存
        for (var i = 0, count = keys.length; i < count; i++) {
            this.remove(keys[i], sid);
        }
    }
}

// 平台默认缓存接口
var storeManager = new MemoryStore();

// 注册自定义缓存接口
exports.register = function(manager) {
    //console.log('暂不支持业务自定义cacheManager缓存接口');
    //return;

    if (typeof manager != 'object') return;
    if (typeof manager.add != 'function') {
        throw new TypeError('注册缓存接口必须包含add法');
    }
    if (typeof manager.get != 'function') {
        throw new TypeError('注册缓存接口必须包含get方法');
    }
    if (typeof manager.remove != 'function') {
        throw new TypeError('注册缓存接口必须包含remove方法');
    }
    console.log('业务自定义cacheManager缓存接口');
    storeManager = manager;
}

// 添加缓存
exports.add = function cacheManager$add(key, value, sid) {
    if (key === null || key === undefined) return;
    storeManager.add(key, value, sid);
}

// 获取缓存
exports.get = function cacheManager$get(key) {
    return storeManager.get(key);
}

// 删除缓存
exports.remove = function cacheManager$remove(key) {
    return storeManager.remove(key);
}

// 清空缓存，慎用
exports.clear = function cacheManager$clear() {
    storeManager.clear();
}

exports.checkExpires = function(req) {
    storeManager.checkExpires(req);
}

