var MySql = require("mysql");
var BaseLib = require("../base/baseLib");
var DebugDBHelper = require('debug')('aprac:DBHelper');
var ApracConfig = require('../../config').apracConfig;

class MySqlLib extends BaseLib {
  constructor(connectionSetting) {
    super(connectionSetting);
  }

  createPool() {
    if (MySqlLib[this.poolName] == null)
      MySqlLib[this.poolName] = MySql.createPool(this.connectionSetting);
  }

  getConnectionFunction() {
    if (this.isPool)
      return MySqlLib[this.poolName].getConnection.bind(MySqlLib[this.poolName]);
    else
      return MySql.createConnection.bind(MySqlLib[this.poolName]);
  }

  async getTableStruct(tablename) {
    let result = await this.query(`SHOW FULL FIELDS FROM ${tablename}`, null);
    if (result.values.length > 0) {
      return result;
    } else {
      throw `未找到表【${tableName}】的结构`;
    }
  }

  async beginTransaction() {
    var _this = this;
    await this.create();
    await new Promise(function (resolver, reject) {
      _this.connection.beginTransaction(function (err) {
        if (err == null) {
          _this.hasTransaction = true;
          resolver();
        }
        else
          reject(err);
      });
    });
  }

  async commitTransaction() {
    var _this = this;
    await this.create();
    await new Promise(function (resolver, reject) {
      _this.connection.commit(function (err) {
        _this.hasTransaction = false;
        if (err == null)
          resolver();
        else
          reject(err);
      });
    });
  }

  async rollbackTransaction() {
    var _this = this;
    await this.create();
    await new Promise(function (resolver, reject) {
      _this.hasTransaction = false;
      if (_this.connection.state == "disconnected") {
        global.setTimeout(function () { resolver(); });
        return;
      }

      _this.connection.rollback(function (err) {
        if (err == null)
          resolver();
        else
          reject(err);
      });
    });
  }

  async query(sql, params) {
    var _this = this;
    try {
      await _this.create();
      return await new Promise(function (resolver, reject) {
        if (typeof sql == "string") {
          sql = { 'sql': sql, timeout: ApracConfig.commandTimeout * 1000 }; // 默认超时时间20秒，配置单位：秒
        } else if (typeof sql == "object") {
          sql.timeout = ApracConfig.commandTimeout * 1000;
        }

        _this.connection.query(sql, params, function (qerr, values, fields) {
          if (qerr == null)
            resolver({
              values: values,
              fields: fields
            });
          else {
            //超时后重新建立数据库连接
            _this._handleDataBaseConnectionError(qerr, function (tempValue) {
              resolver(tempValue);
            }, function (tempErr) {
              reject(tempErr);
            }, sql, params);
          }
        });
      });
    } catch (e) {
      return await new Promise((resolver, reject) => {
        _this._handleDataBaseConnectionError(e, resolver, reject, sql, params);
      })
    }
  }

  close(callback, poolConnectionIsDestroy) {
    super.close(callback, poolConnectionIsDestroy);

    if (this.connection == null) {
      return;
    }

    try {
      if (this.isPool) {
        if (callback) callback.apply(this.connection, arguments);
        if (poolConnectionIsDestroy) this.connection.destroy();
        else this.connection.release();
        this.connection = null;
      } else {
        if (callback == null) {
          this.connection.destroy();
          this.connection = null;
        }
        else {
          var _this = this;
          this.connection.end(function (err) {
            _this._connectionEnd(err);
            if (callback) callback.call(_this, err);
            _this.connection = null;
          });
        }
      }
    } catch (e) {
      DebugDBHelper("关闭连接出错：" + e);
      this.connection = null;
    }
  }

  _connectionConnect(resolver, reject, err) {
    if (err == null) {
      let _this = this;
      this.connection.config.queryFormat = function (query, values) {
        return _this._queryFormat(query, values);
      };
    }

    super._connectionConnect(resolver, reject, err);
  }

  limit(start, count) {
    start = this.sqlEscape(start, null, null, false);
    count = this.sqlEscape(count, null, null, false);
    return ` limit ${start},${count} `;
  }

  toNumber(field) {
    field = this.sqlEscape(field, null, null, false);
    return ` ${field} `;
  }

  IF(condition, truedata, falsedata) {
    condition = this.sqlEscape(condition, null, null, false);
    truedata = this.sqlEscape(truedata, null, null, false);
    falsedata = this.sqlEscape(falsedata, null, null, false);
    return ` IF(${condition},${truedata},${falsedata}) `;
  }

  IFNULL(expression, altValue) {
    expression = this.sqlEscape(expression, null, null, false);
    altValue = (altValue === '' ? "''" : this.sqlEscape(altValue, null, null, false));
    return ` IFNULL(${expression},${altValue}) `;
  }

  SUBSTRING_INDEX(field, split, index) {
    return ` SUBSTRING_INDEX(${field},${split},${index}) `;
  }
}

module.exports = MySqlLib;