var pg = require('pg');
var BaseLib = require('../base/baseLib');
var SqlString = require('sqlstring');
var CHARS_GLOBAL_REGEXP = /[\']/g; // eslint-disable-line no-control-regex
var CHARS_ESCAPE_MAP = {
    '\'': '\'\''
};

class PgSqlLib extends BaseLib {
    constructor(connectionSetting) {
        super(connectionSetting);
    }

    createPool() {
        if (PgSqlLib[this.poolName] == null) {
            PgSqlLib[this.poolName] = pg.Pool({
                host: this.hostname,
                port: this.port,
                user: this.user,
                password: this.password,
                database: this.database,
                poolSize: this.connectionlimit
            });
        }
    }

    getConnectionFunction() {
        if (this.isPool) {
            return PgSqlLib[this.poolName].connect.bind(PgSqlLib[this.poolName]);
        }
        else {
            let client = pg.Client(this.connectionSetting);
            return client.connect.bind(client);
        }
    }

    close(callback, poolConnIsDestroy) {
        super.close(callback, poolConnIsDestroy);

        if (!this.connection) {
            return;
        }

        try {
            if (this.isPool) {
                if (callback) callback.apply(this.connection, arguments);
                // if (poolConnIsDestroy) this.connection.release();
                this.connection.release();
                this.connection = null;
            } else {
                if (callback == null) {
                    this.connection.release();
                    this.connection = null;
                }
                else {
                    var _this = this;
                    this.connection.release(function (err) {
                        _this._connectionEnd(err);
                        if (callback) callback.call(_this, err);
                        _this.connection = null;
                    });
                }
            }
        } catch (e) {
            //DebugDBHelper("关闭连接出错：" + e);
            this.connection = null;
        }
    }

    async query(sql, params) {
        var _this = this;
        if (!this.connection) {
            await _this.create();
        }
        return new Promise(function (resolver, reject) {
            let tsql = _this._queryFormat(sql, params);
            _this.connection.query(tsql, function (qerr, values, fields) {
                if (qerr == null) {
                    _this._pgRowDataFormate(values);
                    resolver({
                        values: values.rows,
                        fields: values.fields
                    });
                }
                else {
                    //超时后重新建立数据库连接
                    _this._handleDataBaseConnectionError(qerr, function (tempValue) {
                        resolver(tempValue);
                    }, function (tempErr) {
                        reject(tempErr);
                    }, sql, params);
                }
            });
        });
    }

    _pgRowDataFormate(values) {//pg数据转换，因为给出来的都是字符串，需要根据类型转数据
        let numberFieldsMap = [];
        let jsonFieldsMap = [];
        if (values.fields != null) {
            values.fields.forEach(field => {
                if (field.dataTypeID == 1043) {//字符串
                }
                if (field.dataTypeID == 20) {//bigint
                    numberFieldsMap.push(field.name);
                }
                if (field.dataTypeID == 25) {//json格式 
                    jsonFieldsMap.push(field.name);
                }
                if (field.dataTypeID == 1700) {//sum
                    numberFieldsMap.push(field.name);
                }
            })
        }

        if ((numberFieldsMap.length > 0 || jsonFieldsMap.length > 0) && values.rows != null) {
            values.rows.forEach(row => {
                numberFieldsMap.forEach(field => {
                    if (row[field]) {
                        row[field] = parseFloat(row[field]);
                    }
                })
                jsonFieldsMap.forEach(jsonfile => {
                    if (row[jsonfile]) {
                        try {
                            row[jsonfile] = JSON.parse(`"${row[jsonfile]}"`)
                        } catch (error) {

                        }
                    }
                });
            })
        }
    }

    async getTableStruct(tableName) {
        let rt = {
            values: []
        };
        if (!tableName) {
            return rt;
        }
        tableName = tableName.toLowerCase();
        let tableStruct = await this.query(`select * from information_schema.columns
        where table_schema='public' and table_name='${tableName}'`, null);
        if (tableStruct.values.length == 0) {
            throw `未找到表【${tableName}】的结构`
        }
        tableStruct.values.forEach(filed => {
            rt.values.push({
                "Field": filed.column_name,
                "Type": filed.udt_name,
                "Collation": "utf8_general_ci",
                "Null": filed.is_nullable,
                "Key": filed.dtd_identifier == "1" ? "PRI" : "",
                "Default": null,
                "Extra": "",
                "Privileges": "",
                "Comment": ""
            })
        })
        tableStruct = rt;

        return tableStruct;
    }

    async beginTransaction() {
        await this.query('BEGIN TRANSACTION;');
        this.hasTransaction = true;
    }

    async rollbackTransaction() {
        await this.query('ROLLBACK;');
        this.hasTransaction = false;
    }

    async commitTransaction() {
        await this.query('COMMIT;');
        this.hasTransaction = false;
    }

    sqlEscape(value, stringifyObjects, timeZone, isAddMake) {
        isAddMake = isAddMake == null ? true : isAddMake;
        let result = this._escape(value, stringifyObjects, timeZone);
        if (isAddMake)
            return result;
        else {
            if (result.startsWith('\'') && result.endsWith('\'')) {
                return result.substr(1, result.length - 2);
            }

            return result;
        }
    }

    _escape(val, stringifyObjects, timeZone) {
        if (val === undefined || val === null) {
            return 'NULL';
        }

        switch (typeof val) {
            case 'boolean': return (val) ? 1 : 0;
            case 'number': return val + '';
            case 'object':
                if (val instanceof Date) {
                    return this._dateToString(val, timeZone || 'local');
                } else if (Array.isArray(val)) {
                    return this._arrayToList(val, timeZone);
                } else if (Buffer.isBuffer(val)) {
                    return this._bufferToString(val);
                } else if (typeof val.toSqlString === 'function') {
                    return String(val.toSqlString());
                } else if (stringifyObjects) {
                    return this._escapeString(val.toString());
                } else {
                    return this._objectToValues(val, timeZone);
                }
            default: return this._escapeString(val);
        }
    }

    _dateToString(date, timeZone) {
        var dt = new Date(date);

        if (isNaN(dt.getTime())) {
            return 'NULL';
        }

        var year;
        var month;
        var day;
        var hour;
        var minute;
        var second;
        var millisecond;

        if (timeZone === 'local') {
            year = dt.getFullYear();
            month = dt.getMonth() + 1;
            day = dt.getDate();
            hour = dt.getHours();
            minute = dt.getMinutes();
            second = dt.getSeconds();
            millisecond = dt.getMilliseconds();
        } else {
            var tz = this._convertTimezone(timeZone);

            if (tz !== false && tz !== 0) {
                dt.setTime(dt.getTime() + (tz * 60000));
            }

            year = dt.getUTCFullYear();
            month = dt.getUTCMonth() + 1;
            day = dt.getUTCDate();
            hour = dt.getUTCHours();
            minute = dt.getUTCMinutes();
            second = dt.getUTCSeconds();
            millisecond = dt.getUTCMilliseconds();
        }

        // YYYY-MM-DD HH:mm:ss.mmm
        var str = "to_timestamp('" +
            (this._zeroPad(year, 4) + '-' + this._zeroPad(month, 2) + '-' + this._zeroPad(day, 2) + ' ' +
                this._zeroPad(hour, 2) + ':' + this._zeroPad(minute, 2) + ':' + this._zeroPad(second, 2))
            + "','yyyy-mm-dd hh24:mi:ss')";

        return str;
    }

    _zeroPad(number, length) {
        number = number.toString();
        while (number.length < length) {
            number = '0' + number;
        }

        return number;
    }

    _convertTimezone(tz) {
        if (tz === 'Z') {
            return 0;
        }

        var m = tz.match(/([\+\-\s])(\d\d):?(\d\d)?/);
        if (m) {
            return (m[1] === '-' ? -1 : 1) * (parseInt(m[2], 10) + ((m[3] ? parseInt(m[3], 10) : 0) / 60)) * 60;
        }
        return false;
    }

    _bufferToString(buffer) {
        return 'X' + this._escapeString(buffer.toString('hex'));
    }

    _objectToValues(object, timeZone) {
        var sql = '';

        for (var key in object) {
            var val = object[key];

            if (typeof val === 'function') {
                continue;
            }

            sql += (sql.length === 0 ? '' : ', ') + SqlString.escapeId(key) + ' = ' + this._escape(val, true, timeZone);
        }

        return sql;
    }

    _arrayToList(array, timeZone) {
        var sql = '';
        for (var i = 0; i < array.length; i++) {
            var val = array[i];

            if (Array.isArray(val)) {
                sql += (i === 0 ? '' : ', ') + '(' + this._arrayToList(val, timeZone) + ')';
            } else {
                sql += (i === 0 ? '' : ', ') + this._escape(val, true, timeZone);
            }
        }
        return sql;
    }

    _escapeString(val) {
        var chunkIndex = CHARS_GLOBAL_REGEXP.lastIndex = 0;
        var escapedVal = '';
        var match;

        while ((match = CHARS_GLOBAL_REGEXP.exec(val))) {
            escapedVal += val.slice(chunkIndex, match.index) + CHARS_ESCAPE_MAP[match[0]];
            chunkIndex = CHARS_GLOBAL_REGEXP.lastIndex;
        }

        if (chunkIndex === 0) {
            // Nothing was escaped
            return "'" + val + "'";
        }

        if (chunkIndex < val.length) {
            return "'" + escapedVal + val.slice(chunkIndex) + "'";
        }

        return "'" + escapedVal + "'";
    }

    limit(start, count) {
        start = this.sqlEscape(start, null, null, false);
        count = this.sqlEscape(count, null, null, false);
        return ` limit ${count} OFFSET ${start} `;
    }

    toNumber(field) {
        field = this.sqlEscape(field, null, null, false);
        return ` to_number(concat('0',${field}), '99999999999999999999') `;
    }

    IF(condition, truedata, falsedata) {
        condition = this.sqlEscape(condition, null, null, false);
        truedata = this.sqlEscape(truedata, null, null, false);
        falsedata = this.sqlEscape(falsedata, null, null, false);
        return ` case when ${condition} then ${truedata} else ${falsedata} end `;
    }

    IFNULL() {
        let arg = arguments;
        let params = '';
        let def = '';
        if (arg.length == 1) {
            params = (arg[0] === '' ? "''" : this.sqlEscape(arg[0], null, null, false));
        } else if (arg.length > 1) {
            def = (arg[arg.length - 1] === '' ? "''" : this.sqlEscape(arg[arg.length - 1], null, null, false));
            for (var i = 0; i < arg.length - 1; i++) {
                params += (arg[i] === '' ? "''" : `${this.sqlEscape(arg[i], null, null, false)},`);
            }
            params = params.substr(0, params.length - 1);
        }
        return ` coalesce(${params}, ${def}) `;
    }

    SUBSTRING_INDEX(field, split, index) {
        return ` split_part(${field},${split},${index}) `;
    }
}
module.exports = PgSqlLib;