﻿/*!
 * apracjs
 * Copyright(c) 2017 by wssf
 * 2017-04-26
 * 处理ajax异步请求
 */
'use strict'
var Handler = require('../handler.js');
var WebClientProxyHandler = require('./webClientProxy.js');
var WebServiceDataJs = require('./webServiceData.js');
var FileUploadHandler = require('./fileUploadHandler.js');
var sysConsts = require('../../scripts/sysConsts.json');
var utils = require('../../common/utils.js');
var loginPage = require('../../scripts/loginPageInfo.js');
var fileUpload = require('../../scripts/fileUpload.js');
const JsonContentType = "application/json";
const UrlEncodedType = "application/x-www-form-urlencoded";

class AjaxHandler extends Handler {
  constructor(app, distributor) {
    super(app, app.get('ajaxTag'), distributor);
  }

  dispose(req) {
    req.methodData = null;
    req.interceptor = null;
    var res = req.res;
    delete req.res;
    delete res.req;

    var method = req.webMethod;
    if (method) {
      method.req = null;
      method.res = null;
      req.webMethod = null;
      if (method.dispose) {
        method.dispose();
      }
    }
  }

  enableSession(req) {
    return !this.isClientProxyRequest(req);
  }

  processRequest(req, res, done) {
    try {
      super.processRequest(req, res, done);
      this._distrributor ? this.distributorInvoke(req, res, done, this) : this.selfInvoke(req, res, done);
    } catch (ex) {
      this.writeExceptionJsonString(req, res, ex);
    }
  }

  distributorInvoke(req, res, done, that) {
    if (this.isClientProxyRequest(req)) {
      this._distrributor.get(req, res, "ajax/js", (values) => {
        WebClientProxyHandler.processRequest(req, res, done, values);
      });
      return;
    }

    var methodName = this.getMethodName(req);
    if (req.method != "POST") {
      throw new Error(utils.format(sysConsts.WebService_InvalidVerbRequest, methodName, "GET"));
    }
    //this.checkLogin(req);//先注释掉
    var ajaxData = {
      methodName: methodName,
      data: req.body
    }
    this._distrributor.get(req, ajaxData, "ajax", (ndata) => {
      var result = {
        value: typeof result == "function" ? ndata : (next) => { next(ndata); }
      }
      that.finishHandler(req, result, that);
    })
  }

  selfInvoke(req, res, done) {
    if (this.isClientProxyRequest(req)) {
      WebClientProxyHandler.processRequest(req, res, done);
      return;
    }

    this.parsedWebMethodData(req);
    this.executeWebServiceCall(req, res, done);
  }

  getMethodName(req) {
    var pathname = req.parsed.pathname;
    var pos = pathname.lastIndexOf(process.platform == "win32" ? "\\" : "/");
    if (pos < 0) {
      throw new Error(sysConsts.WebService_InvalidWebServiceCall);
    }
    var methodName = pathname.substr(pos + 1);
    return methodName;
  }

  parsedWebMethodData(req) {
    var methodName = this.getMethodName(req);
    var serviceData = WebServiceDataJs.getWebServiceData(req);
    var method = serviceData.getMethodData(methodName);
    if (!method) {
      throw new Error(utils.format(sysConsts.WebService_UnknownMethod, methodName));
    }
    req.methodData = method;
    req.interceptor = serviceData.getInterceptor();

    if (req.method != "POST") {
      throw new Error(utils.format(sysConsts.WebService_InvalidVerbRequest, methodName, "GET"));
    }
  }

  async executeWebServiceCall(req, res, done) {
    try {
      await this.checkLogin(req);

      var contentType = req.headers["content-type"];
      if (this.isJsonCallType(contentType) || this.isUrlEncodedType(contentType)) {
        await this.handler(req, req.body);
      }
      else if (this.isMultipart(req)) {
        req.isMultipart = true;
        req.isFlashRequest = this.isFlashRequest(req);
        await this.handler(req, []);
      }
      else {
        throw new Error(utils.format(sysConsts.WebService_InvalidContentType, req.methodData.name, contentType));
      }
    } catch (ex) {
      this.writeExceptionJsonString(req, res, ex);
    }
  }

  async handler(req, data) {
    var res = req.res;
    var that = this;
    this.initCachePolicy(res);

    var result = await this.callMethodFromRawParams(req, data);
    this.finishHandler(req, result, that);
  }

  finishHandler(req, result, that) {
    var res = req.res;
    // 没有返回值的方法
    if (result.value == undefined) {
      this.dispose(req);
      return res.end();
    }

    if (typeof result.value == 'string') {
      this.dispose(req);
      return res.end(result.value);
    }

    // 文件上传
    if (result.value.constructor === fileUpload) {
      return FileUploadHandler.processRequest(result, req, this);
    }

    // 所有回调完成后，结束页面，并输出json
    function successedCallback(data, err) {
      try {
        if (err) {
          that.writeExceptionJsonString(req, res, err);
        } else {
          that.writeJsonString(req, res, data);
        }
      } catch (e) {
        console.log(e);
      }
    };

    // 需要递归处理
    var onlyMethodKey = "_firstMethod";

    // 返回db，返回分页数据源、分页总行数、统计信息等情况，可能存在多个db查询
    var dbFuns = this.getCallbackFuns(result.value, onlyMethodKey);
    var keys = Object.keys(dbFuns);

    // 处理同步回调+同步回调内部的异常
    if (keys.length == 1 && keys[0] == onlyMethodKey) {
      var methodFun = dbFuns[onlyMethodKey];
      if (Object.prototype.toString.call(methodFun) == '[object AsyncFunction]') {
        async function processOn(cb) {
          try {
            await methodFun(cb);
          } catch (e) {
            successedCallback(null, e);
          }
        }
        processOn(successedCallback);
      } else {
        methodFun(successedCallback);
      }
      return;
    }

    async function loopFuns(dbFuns, overBack) {
      if (keys.length == 0) {
        overBack(result.value);
        return;
      }

      var key = keys.splice(0, 1)[0]; // remove

      var fun = dbFuns[key];
      await fun(async (funResult, err) => {
        if (err) {
          overBack(null, err);
          return;
        }
        if (key == onlyMethodKey) {
          overBack(funResult);
          return;
        }

        result.value[key] = funResult; // update the target data

        await loopFuns(dbFuns, overBack); // 递归处理所有promise回调
      });
    };

    // 递归处理所有回调
    loopFuns(dbFuns, successedCallback);
  }

  getCallbackFuns(result, onlyMethodKey) {
    var dbFuns = {};
    if (typeof result == 'function') {
      dbFuns[onlyMethodKey] = result;
      return dbFuns;
    }

    for (var key in result) {
      var fun = result[key];
      if (typeof fun == 'function') {
        dbFuns[key] = fun;
      }
    }
    return dbFuns;
  }

  async checkLogin(req) {
    var answer = await loginPage.checkHasLogin(req, req.methodData.needLogin);
    if (!answer) {
      throw new Error(sysConsts.Session_TimeoutError);
    }
  }

  async callMethodFromRawParams(req, data) {
    var res = req.res;
    var methodData = req.methodData;
    var methodName = methodData.name;
    var methodType = methodData.type;
    var args = methodData.args;
    var parameters = this.strongTypeParameters(req, data, args);

    var classs = new methodType();
    req.webMethod = classs;
    classs.req = req; // 注入依赖
    classs.res = res;
    // req.session = new Proxy(req.session, {
    //     "get": function(target, propertyname) {
    //         if (propertyname !== "req") {
    //             DebugSession("get：" + propertyname);
    //             try {
    //                 if (target[propertyname] === undefined) {
    //                     DebugSession(JSON.stringify(target[propertyname]));
    //                 } else {
    //                     DebugSession(JSON.stringify(target[propertyname]));
    //                 }
    //             } catch (e) {
    //                 DebugSession(e);
    //             }
    //         }
    //         return target[propertyname];
    //     },
    //     "set": function(target, propertyname, value) {
    //         if (propertyname !== "req") {
    //             if (target[propertyname] !== value) {
    //                 DebugSession("set：" + propertyname);
    //                 try {
    //                     if (target[propertyname] === undefined) {
    //                         DebugSession("undefined");
    //                     } else {
    //                         DebugSession(JSON.stringify(target[propertyname]));
    //                     }
    //                 } catch (e) {
    //                     DebugSession(e);
    //                 }
    //             }
    //         }
    //         target[propertyname] = value;
    //     },
    //     "deleteProperty": function(target, propertyname) {
    //         if (target[propertyname] !== undefined) {
    //             DebugSession("delete：" + propertyname);
    //         }
    //         delete target[propertyname];
    //     }
    // });
    classs.session = req.session;

    // Object.defineProperty(classs, 'user', {
    //   get: function() {
    //     return loginPage.userInfo(req);
    //   }
    // });
    classs.user = loginPage.userInfo(req);

    var interceptor = req.interceptor ? new req.interceptor(classs.req) : null;
    let value = await this.wrapInterceptor(interceptor, methodName, classs, parameters);

    return {
      value: value,
      instance: classs
    };
  }

  strongTypeParameters(req, data, args) {
    var parameters = [];
    for (var i = 0, count = args.length; i < count; i++) {
      var arg = args[i];
      if (data[arg] === undefined) {
        throw new Error(utils.format(sysConsts.WebService_NoneParameters, req.methodData.name, arg));
      }
      parameters.push(data[arg]);
    }
    return parameters;
  }

  writeExceptionJsonString(req, res, ex) {
    var status = 500; // 默认异常消息为500
    if (req.isMultipart && req.isFlashRequest) status = 200; // 500的情况，Flash不能获取到异常消息

    res.writeHead(status, {
      'Content-Type': this.getResponseContentType(this.isJsonCall(req)) + "; charset=utf-8",
      'jsonerror': "true"
    });

    this.dispose(req);

    var errorObj = { "Message": utils.makeExceptionHtml(ex, false, req) };
    res.end(JSON.stringify(errorObj));
  }

  writeJsonString(req, res, result) {
    var data = "";
    var contentType = this.getResponseContentType(this.isJsonCall(req));
    if (result !== undefined) {
      if (result instanceof Buffer) {
        //不指定编码格式，方便传输时做数据压缩，同时客户端直接认为传输的是json,会自动序列号
        // contentType = 'text/plain';
        data = result;//输出二进制
      } else {
        data = JSON.stringify(result);
      }
    }
    else {
      data = JSON.stringify({});
    }
    this.resSend(res, data, contentType);
  }

  resSend(res, data, contentType) {
    this.dispose(res.req);

    res.setHeader('Content-Type', contentType);
    res.end(data);
  }

  isClientProxyRequest(req) {
    return req.parsed.pathname.endsWith('\js');
  }

  initCachePolicy(res) {
    res.setHeader("Cache-Control", "private, max-age=0");
  }

  isJsonCallType(contentType) {
    return contentType.startsWith(JsonContentType);
  }

  isUrlEncodedType(contentType) {
    return contentType.startsWith(UrlEncodedType);
  }

  isJsonCall(req) {
    var type = req.headers['content-type'];
    if (type === undefined) type = "";
    return this.isJsonCallType(type);
  }

  isMultipart(req) {
    return req.headers['content-type'].startsWith("multipart/form-data");
  }

  isFlashRequest(req) {
    return req.headers['user-agent'] == 'Shockwave Flash';
  }

  getResponseContentType(jsonCall) {
    return jsonCall ? JsonContentType + ';charset=utf-8' : "text/html";
  }
}

module.exports = AjaxHandler;
