﻿/*!
 * apracjs
 * Copyright(c) 2017 by wssf
 * 2017-04-26
 */

'use strict'
var path = require('path');
var config = require('../../config.js');
var apracConfig = config.apracConfig;
var utils = require('../../common/utils.js');
var StringBuilder = require('../../common/stringBuilder.js');
var StringUtils = require('../../common/stringUtils.js');
var clientUserStore = require('../../scripts/clientUserStore.js');
var IPrinter = require('../../scripts/iPrinter.js');

class PageTag {
  constructor(handler, gspxStr, httpContext) {
    this.handler = handler;
    this.httpContext = httpContext;
    this.req = httpContext.req;
    this.gspxStr = gspxStr;
    this.styles = [];

    this.controlsId = {};
    this.configDatas = {}; // grid\flowpanel\splitter
    this.gridColumnExpandDatas = {};
    this.gridItemCountDatas = {};
  }

  dispose() {
    this.handler = null;
    this.httpContext = null;
    this.req = null;
    this.gspxStr = null;
    this.styles = null;
    this.controlsId = null;
    this.configDatas = null;
    this.gridColumnExpandDatas = null;
    this.gridItemCountDatas = null;
  }

  bindCallbackFuns(dbFuns, data) {
    for (var id in data) {
      var fun = data[id];
      if (typeof fun == "function") {
        fun.__target = data;
        fun.__id = id; // 将id、data、fun三者关联，避免多个hash里面出现相同id的情况
        dbFuns.push(fun);
      }
    }
  }

  renderPage() {
    var dbFuns = [];
    this.bindCallbackFuns(dbFuns, this.httpContext.getContextData());    // 上下文绑定的数据源
    this.bindCallbackFuns(dbFuns, this.gridItemCountDatas);              // 数据总行数
    this.bindCallbackFuns(dbFuns, this.configDatas);                     // grid\flowpanel\splitter的配置数据

    if (dbFuns.length == 0) return this.successedCallback();             // 如果没有fun待执行，直接输出页面

    var overBack = utils.createDelegate(this, this.successedCallback);
    loopFuns(dbFuns);

    function loopFuns(dbFuns) {
      var fun = dbFuns.pop();
      fun((result, ex) => {
        if (ex) {
          return overBack(ex);                                          // 回调直接抛出异常
        }
        fun.__target[fun.__id] = result;                                // 替换fun为真实计算后的值

        if (dbFuns.length == 0) return overBack();

        loopFuns(dbFuns);                                               // 递归处理所有promise回调
      });
    };
  }

  async successedCallback(ex) {
    var req = this.req;
    if (ex) {
      return this.endResponse(req.res, utils.makeExceptionHtml(ex, null, req));
    }
    var pageInfo = this.httpContext.pageInfo;

    var sb = new StringBuilder();
    this.doPagePart("configDatas", JSON.stringify(this.configDatas), sb);
    this.doPagePart("itemCountDatas", JSON.stringify(this.gridItemCountDatas), sb);
    this.doPagePart("gridColumnExpandDatas", JSON.stringify(this.gridColumnExpandDatas), sb);
    this.doPagePart("itemCountDatas", '{}', sb);
    this.doPagePart("pageId", this.getID(), sb);
    this.doPagePart("internalID", this.getInternalID(), sb);
    this.doPagePart("rootUrl", '', sb);

    this.doPagePart("apracVersion", apracConfig.versionTag, sb);

    var v = pageInfo.getScriptVersion();
    this.doPagePart("version", v, sb);
    this.doPagePart("scriptIncludePage", `R.axd?v=${v}&r=`, sb);

    this.doPagePart("scriptPath", req.parsed.path.replace('\\', '/'), sb);
    this.doPagePart("convertUrl", apracConfig.webFolder, sb);
    this.doPagePart("hasController", pageInfo.hasController().toString(), sb);
    this.doPagePart("hasWebService", pageInfo.hasWebService().toString(), sb);

    if (pageInfo.hasWebService() && pageInfo.hasController()) {
      this.doPagePart("controllerName", pageInfo.controllerName, sb);
      this.doPagePart("webMethodBasePath", pageInfo.webMethodBasePath(), sb);
    }

    this.doPagePart("reportSession", JSON.stringify(IPrinter.getInitData(req.session)), sb);

    this.doPagePart("baseUrl", this.getBaseUrl(req.url), sb);
    if (req.httpContext.getIsAsyncRequest()) { // 嵌入MDI方式，必须指定variables， 由于没有做后期拼接，所以不会受转义字符影响
      this.doPagePart("variables", this.httpContext.getContextDataString(), sb);
    }
    this.doPagePart("gspx", this.gspxStr ? this.gspxStr : '', sb);

    var html;
    if (!req.httpContext.getIsAsyncRequest()) {
      if (IPrinter.reportConfig) {
        var reportConfig = {};
        await IPrinter.reportConfig.Init(req, reportConfig);
        config.apracSettings.ReportConfig = reportConfig;
      }
      html = sb.toString();
      html = this.renderHTML(html, JSON.stringify(config.apracSettings), this.httpContext.getContextDataString(), req);
    }
    else if (req.httpContext.getHeaderValue('getsettings') === '1') {
      this.doPagePart("settings", JSON.stringify(config.apracSettings), sb);
      html = sb.toString();
    }
    else {
      html = sb.toString();
    }

    this.endResponse(req.res, html);
  }

  endResponse(res, html) {
    this.handler.endResponse(res, html);
    this.dispose();
  }

  getBaseUrl(url) {
    var index = url.indexOf('?');//Login.gspx?Redir=Inner\Default.gspx处理这种URL
    var url = path.dirname(index >= 0 ? url.substr(0, index) : url);
    return url.endsWith("/") ? url : url + "/";
  }

  renderHTML(pageParts, apracSettings, variables, req) {
    var html = new StringBuilder();
    html.appendLine(`<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">`);
    html.appendLine(`<html xmlns= "http://www.w3.org/1999/xhtml">`);
    html.appendLine(`<head>`);
    html.appendLine(`  <title>...</title>`);
    html.appendLine(`  <meta http-equiv="Content-Type" content="text/html; charset=utf8"/>`);
    html.appendLine(`  <meta http-equiv="X-UA-Compatible" content="IE=edge"/>`);
    html.appendLine(`  <link rel="shortcut icon" type="image/x-icon" href="favicon.ico"/>`);


    this.appendStyle(html, 'aprac/skins/craba.min.css', apracConfig.versionTag);

    this.appendConfigStyles(html, apracConfig.loadCss, apracConfig.productVersionTag);

    if (!StringUtils.isNullOrEmpty(apracConfig.defaultSkinName)) {
      this.appendStyle(html, `skins/${apracConfig.defaultSkinName}/skin.css`, apracConfig.productVersionTag);
    }

    this.appendOtherStyles(html, apracConfig.productVersionTag);

    html.appendLine(`</head>`);
    if (apracConfig.defaultSkinName) html.appendLine(`<body class="${apracConfig.defaultSkinName}">`);
    html.appendLine(`  <div id="LoadingFirst" class="Loading"><span class="iconfont LoadingIcon"></span>正在加载...</div>`);

    this.appendReourceScript(html, 'aprac/js/craba.min.js', apracConfig.versionTag);
    this.appendReourceScript(html, 'aprac/js/crabaEx.min.js', apracConfig.versionTag);

    //this.appendReourceScript(html, 'aprac/js/content.min.js', apracConfig.versionTag); // 没有用到它，少一个请求
    //this.appendReourceScript(html, 'aprac/js/echarts-all.js', apracConfig.versionTag); // 前端动态加载的

    this.appendConfigScripts(html, apracConfig.loadScripts, apracConfig.productVersionTag);

    html.appendLine(`  <script>`);
    html.appendLine(`     var __crabaResourceURI = "${apracConfig.virtualPath}/R.axd?v=${apracConfig.versionTag}&r=aprac/";`);
    html.appendLine(`     var __webResourceURI = "${apracConfig.virtualPath}/R.axd?v=${apracConfig.versionTag}&r=/";`);
    html.appendLine(`     var __skinBaseURI = "${apracConfig.virtualPath}/R.axd?v=${apracConfig.versionTag}&r=aprac/skins/";`);
    html.appendLine(`     var __skinName = "${apracConfig.defaultSkinName}";`);
    html.appendLine(`     var __skinPath = "${apracConfig.virtualPath}/R.axd?v=${apracConfig.versionTag}&r=aprac/skins/${apracConfig.defaultSkinName}/";`);
    html.appendLine(`     var __rootURI = "${apracConfig.virtualPath}/";`);
    html.appendLine(`     var __isStandard = 1; var __pathMapName='';`);
    this.initClientUserStore(html, req);
    html.appendLine(`     $app.initialize(${apracSettings});`);
    html.appendLine(`     $app.add_init(function $xxx$__init() {`);
    html.appendLine(`         var form = $app.getForm("${this.getID()}", false);`);
    html.appendLine(`         var pageEx = new Sys.Net.PageRequestEx(true);`);
    html.appendLine(`         var gspxString = '${pageParts}';`);
    html.appendLine(`         var delta = pageEx._parseReplyClient(gspxString, true, ${variables});`);
    html.appendLine(`         if (delta.error) {`);
    html.appendLine(`             var error = Sys.Net.WebServiceProxy.formatWebServiceFailedServerError(delta.error);`);
    html.appendLine(`             $common.alert(error);`);
    html.appendLine(`         } else {`);
    html.appendLine(`             var delegateSuccess = Function.createDelegate(form, form.serverResolvePageLoaded);`);
    html.appendLine(`             pageEx._doSuccessClient(delegateSuccess, delta);`);
    html.appendLine(`         }`);
    html.appendLine(`     });`);
    html.appendLine(`  </script>`);

    this.appendConfigScripts(html, apracConfig.afterLoadScripts, apracConfig.productVersionTag);

    html.appendLine(`</body>`);
    html.appendLine(`</html>`);

    return html.toString();
  }

  initClientUserStore(html, req) {
    var store = clientUserStore.getStore()
    if (store && store.inited) {
      var data = {};
      var keys = Object.keys(store.options);
      for (var i in keys) {
        var key = keys[i];
        var value = store.options[key];
        data[key] = value;
        data[value] = req.session[value] !== undefined ? req.session[value] : null;
      }
      html.appendLine(`     window.$user = ${JSON.stringify(data)};`);
    }
  }

  getID() {
    if (this._id) return this._id;  // 相同页面，不同查询参数将有不同的 ID

    var postfix = this.req.httpContext.get_pageParamsString();
    var query = this.req.parsed.query;
    if (!postfix) {
      postfix = JSON.stringify(query);
    }
    if (!postfix || postfix == '{}') {
      postfix = '';
    }

    var pageInfo = this.httpContext.pageInfo;
    this._id = this.getHashID(pageInfo.controllerName + postfix); // 这个id的生成规则不能随意改动，否则页面的配置及相关历史数据保存全丢失
    return this._id;
  }

  getHashID(url) {
    return utils.getHashCode(url);
  }

  getControlID(profix) {
    var sequence;
    var temp = this.controlsId[profix];
    if (!temp) {
      sequence = 1;
    } else {
      sequence = parseInt(temp) + 1;
    }
    this.controlsId[profix] = sequence;
    return profix + sequence; // 与服务端要统一
  }

  getControlClientID(id) {
    var answer = id;
    if (id && id.trim().length > 0) {
      answer = "$" + this.getID() + "$" + id;
    }
    return answer;
  }

  getInternalID() {
    if (this._internalId != null) {
      return this._internalId;
    }

    var pageInfo = this.httpContext.pageInfo;
    return this._internalId = this.getHashID(pageInfo.controllerName);
  }

  doPagePart(partName, partValue, buffer) {
    buffer.appendFormat("{0}|{1}|{2}|", partValue.length, partName, partValue);
  }

  appendStyle(html, r, v) {
    html.appendFormat('  <link type="text/css" rel="stylesheet" href="R.axd?v={0}&r={1}"/>', v, r).appendLine();
  }

  appendOtherStyles(html, v) {
    // 预先加载首页css，避免前端解析布局时延迟渲染问题
    for (var i = 0, count = this.styles.length; i < count; i++) {
      var style = this.styles[i];
      var id = this.getID() + "_index" + i;
      if (style.Url) {
        html.appendFormat('  <link type="text/css" id="{2}" rel="stylesheet" href="R.axd?v={0}&r={1}"/>', v, style.Url, id).appendLine();
      } else if (style.Src) {
        html.appendFormat('  <link type="text/css" id="{2}" rel="stylesheet" href="{1}?v={0}"/>', v, style.Src, id).appendLine();
      } else if (style.Value) {
        html.appendFormat('  <style type="text/css" id="{0}">{1}</style>', id, style.Value).appendLine();
      }
    }
  }

  appendConfigStyles(html, cssString, v) {
    if (!cssString) return;
    var arr = cssString.toString().split(';');
    for (var i in arr) {
      var css = arr[i];
      if (!css.trim()) continue;

      if (css.startsWith('http')) {
        html.appendFormat('  <link type="text/css" rel="stylesheet" href="{0}?v={1}"/>', css.trim(), v).appendLine();
      } else {
        html.appendFormat('  <link type="text/css"  rel="stylesheet" href="R.axd?v={0}&r={1}"/>', v, css.trim()).appendLine();
      }
    }
  }

  appendConfigScripts(html, scriptString, v) {
    if (!scriptString) return;
    var arr = scriptString.toString().split(';');
    for (var i in arr) {
      var script = arr[i];
      if (!script.trim()) continue;

      if (script.startsWith('http')) {
        html.appendFormat('  <script type="text/javascript" src="{0}?v={1}" ></script>', script.trim(), v).appendLine();
      } else {
        html.appendFormat('  <script type="text/javascript" src="R.axd?r={0}&v={1}&" ></script>', script.trim(), v).appendLine();
      }
    }
  }

  appendReourceScript(html, r, v) {
    html.appendFormat('  <script type="text/javascript" src="R.axd?v={0}&r={1}" ></script>', v, r).appendLine();
  }

  appendIncludeScript(html, src) {
    html.appendFormat('  <script type="text/javascript" src="{0}" ></script>', src).appendLine();
  }

  addStyle(style) {
    this.styles.push(style);
  }

  addConfigData(id, configData) {
    this.configDatas[id] = configData;
  }

  addGridColumnExpandDatas(id, expandDatas) {
    this.gridColumnExpandDatas[id] = expandDatas;
  }

  addGridItemCountDatas(id, countDatas) {
    this.gridItemCountDatas[id] = countDatas;
  }
}

module.exports = PageTag;