﻿/*!
 * ConfigStore配置存储器
 * Copyright(c) 2017 by wssf
 * 2017-06-29
 */

'use strict'
var urlParse = require('url').parse;
var debug = require('debug')('aprac:configStore');
var config = require('../config.js');
var apracConfig = config.apracConfig;
var log = require('../log/log4Helper.js').getInstance();
var utils = require('../common/utils.js');
var DbHelper = require('../database/dbHelper.js');
var SqlHelper = require('../database/sqlHelper.js');

var storeProvider = null;
var configCache = {}; // 缓存配置数据，避免频繁读取数据库

exports.registerStoreProvider = function(provider) {
  // 产品可以自定义配置管理器
  storeProvider = provider;
}

exports.saveData = function configStore$saveData(req, id, data, typeName) {
  // 调用配置管理器存放配置
  return getProvider().saveData(req, id, data, typeName);
};

exports.loadData = async function configStore$saveData(req, id, typeName) {
  // 调用配置管理器读取配置
  return await getProvider().loadData(req, id, typeName);
};

exports.init = function(options) {
  // 预先初始化配置器，避免异步问题，读写配置时数据库还没初始化的现象
  getProvider().init(options);
}


// 私有方法
function getProvider() {
  // 获取配置管理器，如果产品没有注册，则使用平台默认的配置管理器
  if (storeProvider) return storeProvider;
  return storeProvider = new ApracStoreProvicer();
}

// 私有，平台默认配置管理器
class ApracStoreProvicer {
  constructor() {
    this.options = {
      loginUserIdKey: "LoginUserId",
      profileIdKey: "ProfileId",
      connNameKey: "ConfigConn",
      useStringKey: false // 业务通过newCuidString() 获取的id为字符串，按需配置，默认为newCuid()整数类型；
    }

    this.dbConnStr = null;
    this.dbError = false;
    this.dbInited = false;
    this.database = null;
    this.writeCount = 0;
    this.maxErrorCount = 20;
    this.tablename = "configstore";
  }

  async saveData(req, id, data, typeName) {
    if (this.dbError) return "db error";
    if (!this.dbInited) return "init faild";

    var profileId = req.session[this.options.profileIdKey];
    var userId = req.session[this.options.loginUserIdKey];
    var configId = typeName + '-' + id;
    var keyId = this.getKeyId(profileId, userId, configId);

    var db = this.createDbHelper();

    try {
      var result = await this.deleteConfigData(db, profileId, userId, configId);

      if (!data) {
        configCache[keyId] = '';
        return 'clear ok';
      }

      var text = JSON.stringify(data);
      configCache[keyId] = data;
      var params = {
        userid: userId,
        configid: configId,
        text: text,
        updatetime: new Date()
      };

      var sql = "insert into configstore (userid,configid,text,updatetime) values (:userid,:configid,:text,:updatetime)";
      if (profileId) {
        params.profileid = profileId;
        sql = "insert into configstore (profileid,userid,configid,text,updatetime) values (:profileid,:userid,:configid,:text,:updatetime)";
      }
      result = await db.executeSQL(sql, params);
      return 'insert ok';
    } catch (ex) {
      configCache[keyId] = '';
      return 'insert error:' + ex.message;
    } finally {
      await db.close();
    }
  }

  getKeyId(profileId, userId, configId) {
    if (!profileId) profileId = 0;
    return profileId + '-' + userId + '-' + configId;
  }

  async loadData(req, id, typeName) {
    if (this.dbError || !this.dbInited) return null;

    var profileId = req.session[this.options.profileIdKey];
    var userId = req.session[this.options.loginUserIdKey];
    var configId = typeName + "-" + id;
    var keyId = this.getKeyId(profileId, userId, configId);

    var data = configCache[keyId]; // 取缓存
    if (data !== undefined) {
      return data;
    }

    var db = this.createDbHelper();
    try {
      var item = await this.getConfigData(db, profileId, userId, configId); // 取数据库配置

      if (!item) {
        configCache[keyId] = ''; // 标志已经读取过，避免重复读取
      } else {
        data = JSON.parse(item.text);
        configCache[keyId] = data; // 存入缓存
      }
    } catch (ex) {
      configCache[keyId] = '';
    } finally {
      await db.close();
    }
    return configCache[keyId];
  }

  async deleteConfigData(db, profileId, userId, configId) {
    var params = {
      userid: userId,
      configid: configId
    };
    var sql = "delete from configstore where userid=:userid and configid=:configid";
    if (profileId) {
      sql += " and profileid=:profileid";
      params.profileid = profileId;
    }

    var result = await db.executeSQL(sql, params);
    return result.values.affectedRows > 0;
  }

  async getConfigData(db, profileId, userId, configId) {
    var params = {
      userid: userId,
      configid: configId
    };
    var sql = "select profileid,userid,configid,text from configstore where userid=:userid and configid=:configid";
    if (profileId) {
      sql += " and profileid=:profileid";
      params.profileid = profileId;
    }

    var result = await db.select(sql, params);
    return result.values[0];
  }

  init(options) {
    log.writeInfo('configStore init.', debug);
    utils.copyTo(options, this.options);

    this.dbConnStr = apracConfig[this.options.connNameKey];
    this.dbError = !this.dbConnStr;
    this.dbInited = false;

    if (this.dbError) {
      var info = "ConfigStore初始化失败，" + this.options.connNameKey + "为空";
      log.writeInfo(info, debug);
      return;
    }

    this.checkDataBase();
  }

  async checkDataBase() {
    var url = urlParse(this.dbConnStr, true);
    var database = this.database = url.pathname.substr(1);
    var serverConn = this.dbConnStr.replace(database, 'information_schema').replace('useConnectionPooling', ''); //  必须替换掉，否则会冲突
    var db = this.createDbHelper(serverConn);

    try {
      var result = await SqlHelper.databaseExists(db, database);
      if (result) {
        return await this.checkTable();
      }

      result = await SqlHelper.createDatabase(db, database);
      if (result) {
        return await this.checkTable();
      }
    } catch (ex) {
      this.createError("ConfigStore初始化失败，创建配置数据库无效", ex.message);
    } finally {
      await db.close();
    }
  }

  async checkTable() {
    var db = this.createDbHelper();
    try {
      var result = await SqlHelper.tableExists(db, this.database, this.tablename);
      if (result) {
        return this.initEnd();
      }

      result = await SqlHelper.createTables(db, this.getTableSqls());
      if (result) {
        return this.initEnd();
      }
    } catch (ex) {
      this.createError("ConfigStore初始化失败，创建配置数据表无效", ex.message);
    } finally {
      await db.close();
    }
  }

  initEnd() {
    this.dbInited = true;
    this.dbError = false;
    log.writeInfo("------ConfigStore初始化成功---------", debug);
  }

  getTableSqls() {
    var sqls = [];

    if (this.options.useStringKey) {
      var sql1 = `create table configstore(
      profileid varchar(30) default '0',
      userid varchar(30) not null default '0' ,
      configid varchar(100) NOT NULL,
      text text,
      updatetime datetime,
      PRIMARY KEY (profileid, userId, configId)
      ) ENGINE = MyISAM DEFAULT CHARSET=utf8`;
      sqls.push(sql1);
    } else {
      var sql1 = `create table configstore(
      profileid bigint unsigned default 0,
      userid bigint unsigned not null default 0 ,
      configid varchar(100) NOT NULL,
      text text,
      updatetime datetime,
      PRIMARY KEY (profileid, userId, configId)
      ) ENGINE = MyISAM DEFAULT CHARSET=utf8`;
      sqls.push(sql1);
    }

    return sqls;
  }

  createDbHelper(conn) {
    return new DbHelper(conn ? conn : this.dbConnStr, null);
  }

  createError(result, err) {
    this.dbError = true;
    log.writeInfo("ConfigStore初始化失败" + err, debug);
  }

  writeError(result, err) {
    if (!err) return; // 正常的回调，ok()后的回调

    this.writeCount++;
    debug(err);

    if (this.writeCount > this.maxErrorCount) {
      this.dbError = true;
      log.writeInfo("ConfigStore 操作失败:" + this.writeCount + "次，大于最大限制次数，配置管理器自动关闭", debug);
    } else {
      log.writeInfo("ConfigStore 操作失败:" + this.writeCount + "次", debug);
    }
  }
}