﻿/*!
 * PagerDataSource
 * Copyright(c) 2017 by wssf
 * 2017-06-09
 */

'use strict'
var PagerDataSource = require('./pagerDataSource.js');
var debug = require('debug')('aprac:SqlPagerDataSource');
var utils = require('../../common/utils.js');
var DbHelper = require('../../database/dbHelper.js');
var config = require('../../config.js').apracConfig;
var StringBuilder = require('../../common/stringBuilder.js');
var stringUtils = require('../../common/stringUtils.js');

class SqlPagerDataSource extends PagerDataSource {
  constructor(connectionString) {
    super();
    this.connectionString = connectionString;
  }

  setQueryParams(queryParams) {
    super.setQueryParams(queryParams);
    this.invalidateCache();
  }

  invalidateCache() {
    super.invalidateCache();
    this._itemCount = -1;
    this._sumCache = null;
  }

  getItemList(first, count) {
    // 取分页数据列表
    var _this = this;

    return async (cb) => {
      var db = new DbHelper(_this.connectionString, null);//统一采用数据库连接池
      try {
        var pagerParams = { 'first': first, 'count': count, 'orderClause': '', 'limitSelectClause': '' };

        await _this.internalPrepare(db, pagerParams);

        var sql = _this.internalLimitSql(db, pagerParams);
        var result = await db.select(sql, db.getParameters());
        db.close();

        var list = await _this.beforeGetItemList(result.values);
        if (!list) {
          throw Error('beforeGetItemList必须返回数据集');
        }

        if (_this.afterSelect) {
          db = new DbHelper(_this.connectionString, null);//统一采用数据库连接池
          await _this.afterSelect(db, pagerParams, list); // 子类重写，比如SqlTreeGridDataSource
          db.close();//finally方法不会立即执行，必须立即回收，如果连接池满了后造成死锁
        }

        cb(list);
      }
      catch (ex) {
        db.close();
        cb(null, ex);
      }
    }
  }

  async internalPrepare(dbHelper, pagerParams) {
    await this.doPrepare(dbHelper, pagerParams);
    if (!pagerParams.selectClause && !pagerParams.limitSelectClause)
      throw new Error("缺少参数SelectClause");
  }

  async beforeGetItemList(list) {
    return list;
  }

  getItemCount() {
    // 取当前数据源总记录条数
    if (this._itemCount > 0) return this._itemCount; // 返回缓存的记录数，翻页的时候不用重复计算，仅查询invalidateCache后才执行

    var that = this;

    return async (cb) => {
      var db = new DbHelper(this.connectionString, null);//统一采用数据库连接池
      var pagerParams = { 'first': 0, 'count': 0 };
      try {
        await that.doSum(db);
        if (that._itemCount < 0) {
          await this.doPrepareCount(db, pagerParams); // 业务重写自行返回合计sql，性能更优
          var sql = pagerParams.selectClause;
          if (!sql) {
            await that.internalPrepare(db, pagerParams);
            sql = "select count(*) as c from (" + pagerParams.selectClause + ") __a"; // 平台按查询sql计算合计性能差，建议产品重写doPrepareCount
          }

          let result = await db.executeScalerSQL(sql);
          that._itemCount = result;
          db.close();//finally方法不会立即执行，必须立即回收，如果连接池满了后造成死锁
          cb(result);
        } else {
          db.close();//finally方法不会立即执行，必须立即回收，如果连接池满了后造成死锁
          cb(that._itemCount);
        }
      }
      catch (ex) {
        db.close();
        cb(null, ex); // 跟getItemList一样，要输出异常
      }
    }
  }

  set_itemCount(v) {
    this._itemCount = v;
  }

  set_sumData(v) {
    this._sumCache = v;
  }

  getSumData() {
    if (this._sumCache != null) return this._sumCache;

    return {};
  }

  async doSum(db) {
    // 统计合计+取记录数，子类处理
  }

  doPrepareCount(db, pagerParams) {
    // 计算记录数，子类处理
  }

  doPrepare(dbHelper, pagerParams) {
    // 计算分页数据，子类处理
  }

  addSummaryItems(items) {
    // 计算合计，子类处理
  }

  internalLimitSql(dbHelper, pagerParams) {
    var sql = pagerParams.limitSelectClause;
    if (sql) return sql;

    var first = pagerParams.first;
    var count = pagerParams.count;
    var orderClause = pagerParams.orderClause;
    if (orderClause != null && !orderClause.startsWith(' ')) {
      pagerParams.orderClause = ' ' + orderClause;
    }
    sql = pagerParams.selectClause + pagerParams.orderClause;
    if (!this._showAll && count >= 0) {
      sql += dbHelper.limit(first, count)
    }
    return sql;
  }

  getFilterClause(dbHelper, prefix) {
    if (this.filter == null) return "";
    if (!prefix) prefix = "";

    this.invalidateCache();

    if (this.filter.items.length == 0) return "";

    return utils.format("{0}{1}", prefix, this.filterToString(dbHelper)); // and filter 或 where filter
  }

  filterToString(db) {
    var rowFilter = new StringBuilder();
    var items = this.filter.items;
    for (var i = 0, count = items.length; i < count; i++) {
      var item = items[i];
      if (!this.isValid(item)) continue;

      var sqlExpr = this.getItemSqlExpr(db, item);

      if (rowFilter.length() > 0) {
        rowFilter.append(" AND ");
      }
      rowFilter.appendFormat("({0})", sqlExpr);
    }
    return rowFilter.toString();
  }

  isValid(item) {
    var type = item.type;
    return type != 1 ?
      (type == 2 ? true : !stringUtils.isNullOrEmpty(item.value)) :
      (!stringUtils.isNullOrEmpty(item.value1) || !stringUtils.isNullOrEmpty(item.value2));
  }

  getItemSqlExpr(db, item) {
    var fieldName = item.dataField; // SqlHelper.FilterField(dataField); // 安全起见，来自外部的拼接字符，必须做检查且参数化值
    var value = item.value;
    var value1 = item.value1;
    var value2 = item.value2;

    switch (item.type) {
      case 0:
        db.addLikeParameter(fieldName, value);
        return utils.format("{0} LIKE :{0}", fieldName);

      case 1:
        var ok1 = !stringUtils.isNullOrEmpty(value1);
        var ok2 = !stringUtils.isNullOrEmpty(value2);
        var key1 = utils.format("{0}_1", fieldName);
        var key2 = utils.format("{0}_2", fieldName);

        if (ok1 && ok2) {
          db.addParameter(key1, value1);
          db.addParameter(key2, value2);
          return utils.format("{0}>=:{1} AND {0}<=:{2}", fieldName, key1, key2);
        }
        if (ok1) {
          db.addParameter(key1, value1);
          return utils.format("{0}>=:{1}", fieldName, key1);
        }
        else {
          db.addParameter(key2, value2);
          return utils.format("{0}<=:{1}", fieldName, key2);
        }

      case 2:
        db.addParameter(fieldName, value);
        var answer = utils.format("{0}=:{0}", fieldName);
        if (stringUtils.isNullOrEmpty(value)) {
          answer += utils.format(" or {0} is null", fieldName);
        }
        return answer;

      default:
        throw new Error("type " + type + "类型错");
    }
  }
}

module.exports = SqlPagerDataSource;